local CLUSTERMETHODS = {};

local function DEBUG(msg) 
   DEFAULT_CHAT_FRAME:AddMessage("SRUC: " .. msg);
end

function CLUSTERMETHODS:ApplyLocationOptions(opts)
   local parent = self:GetParent();
   local pHeight = parent:GetHeight();
   local pWidth = parent:GetWidth();
   local locX = (opts.locX or 0) / 2.0;
   local locY = (opts.locY or 0) / 2.0;
   local mHeight = self:GetHeight();
   local scale = (pHeight / mHeight) * (opts.scale or 0.5);
   self:SetScale(scale);
   self:ClearAllPoints();
   self:SetPoint("CENTER", parent:GetName(), "CENTER", 
		 (pWidth * locX) / scale,
		 (pHeight * locY) / scale);
end

function CLUSTERMETHODS:ApplyUnitConfigMap(configSet, noSticky)
   local sType = type(unitSpec);
   
   for _,frame in ipairs(self.unitFrames) do
      local unit = frame.unit;
      local config = configSet[unit];
      if (config) then
	 frame:ApplyConfig(config, notSticky);
      end
   end
end

function CLUSTERMETHODS:NotifyInfoUpdate(model)
   for _,frame in ipairs(self.unitFrames) do
      frame:NotifyInfoUpdate(visState);
   end
end

function CLUSTERMETHODS:PreviewInfoModel(unitSpec, model, fallbackModel)
   local sType = type(unitSpec);
   
   for _,frame in ipairs(self.unitFrames) do
      local doApply;
      if (sType == "table") then
	 doApply = unitSpec[frame.unit];
      elseif (sType == "string") then
	 doApply = (unitSpec == frame.unit);
      elseif (sType == "function") then
	 doApply = unitSpec(frame.unit);
      end

      if (doApply) then
	 frame:SetInfoModel(model, false, true);
      elseif (fallbackModel) then
	 frame:SetInfoModel(fallbackModel, false, true);
      end
   end
end

function CLUSTERMETHODS:RestoreInfoModel()
   for _,frame in ipairs(self.unitFrames) do
      frame:SetInfoModel(nil);
   end
end

function CLUSTERMETHODS:ApplyScaffold(scaffoldSpec)
   -- TODO Deal with changing this at some point
   local scaffold = StatRingsScaffold_New();
   this.scaffold = scaffold;
   this.scaffoldSpec = scaffoldSpec;
   local _, err = scaffold:Compile(scaffoldSpec, nil, true);

   if (err) then
      DEBUG("Failed to compile scaffold: " .. err);
      return;
   end

   scaffold:SetRootLocation("root", 0, 0, 100, 0);

   for _,frame in ipairs(self.unitFrames) do
      local unit = frame.unit;
      if (scaffold:CanSetActive(unit)) then
	 scaffold:SetActive(unit, frame.visible);
      end
      frame:ApplyScaffoldAnchor(scaffold, unit, 
				scaffold:GetLocation(unit));
   end

   scaffold:Update(); -- Apply changes

   for _,frame in ipairs(self.unitFrames) do
      -- DEBUG("Applying anchor to " .. frame.unit);
      local unit = frame.unit;
      frame:ApplyScaffoldAnchor(scaffold, unit, 
				scaffold:GetLocation(unit));
      scaffold:Attach(unit, frame, "ApplyScaffoldAnchor");

   end
   scaffold:Update();
end

function StatRingsCluster_OnLoad(this, unitFrames)
   if (not unitFrames) then
      message("StatRingsCluster_OnLoad called without any unitFrames");
   end
   this.unitFrames = unitFrames;

   local newMeta = {};
   local oldMeta = getmetatable(this);
   for k,v in pairs(oldMeta) do newMeta[k] = v; end
   function newMeta.__index(table, index) 
      return oldMeta.__index(table,index) or CLUSTERMETHODS[index];
   end
   setmetatable(this, newMeta);
end
