--STAT_RINGS_TEST_UNIT="player";
--STAT_RINGS_TEST_PET="pet";

local function DEBUG(msg)
   DEFAULT_CHAT_FRAME:AddMessage("SRUF: " .. msg);
end

local METHODS = {};

local RINGLAYOUT_SPLIT = {
   SetupRings =
      function(self, frame)
	 local rotation = frame.rotation or 0;

	 frame.manaFrame:SetScale(1.0);
	 frame.manaFrame:SetReversed(true);
	 frame.manaFrame:SetAngle(self:GetManaAngle(0, rotation));

	 frame.healthFrame:SetScale(1.0);
	 frame.healthFrame:SetReversed(false);
	 frame.healthFrame:SetAngle(self:GetHealthAngle(0, rotation));
      end,

   maxAngle = 120,
   startAngle = 210,
   
   GetHealthAngle = 
      function(self, health, rotation) 
	 return self.maxAngle * health, self.startAngle + rotation
      end,

   GetManaAngle = 
      function(self, mana, rotation) 
	 return self.maxAngle * mana, self.startAngle - rotation; 
      end,
}

-- Make sure ALL of the configuratoin settings are in here
local RINGCONFIG_STANDARD = {
   alphaFull = 0,
   timeFull = 4,
   alphaAlert = 0.7,
   timeAlert = 1,
   healthAlert = 0.5,
   manaUnfade = 1,
   alphaNormal = 0.5,
   timeNormal = 1,
}

function StatRingsUnitFrame_SetRadialPosition(frame, angle, offset, parent)
   if (not parent) then
      parent = frame:GetParent();
      if (parent) then parent = parent:GetName(); end;
   end
   local rad = math.rad(angle);
   frame:SetPoint("CENTER", parent, "CENTER",
		  math.sin(rad) * offset, math.cos(rad) * offset);
end

METHODS.SetRadialPosition = StatRingsUnitFrame_SetRadialPosition;

function METHODS:HandleShow()
   -- If we've been asked to show but are configured not to
   -- then hide again
   if (not self.visible) then
      self:Hide();
      return;
   end
   if (self.requiresRefresh) then
      self.instantFade = true;
      self:NotifyInfoUpdate();
   end
end

-- OnUpdate handler for fading
local function DoFade()
   if ((not this.fading) or (not this.fadeGoal)) then
      -- DEBUG("Cancelling fade");
      this.fadeStart = nil;
      this.fading = nil;
      this.fadeGoal = nil;
      this:SetScript("OnUpdate", nil);
      return;
   end
   local now = GetTime();
   local elapsed = now - this.fadeStart;
   if (arg1 < elapsed) then
      elapsed = arg1;
   end
   this.fadeStart = now;
   local motion = this.fading * elapsed;
   local newLevel = this.curAlpha + motion;
   if (motion < 0) then
      if (newLevel < this.fadeGoal) then
	 newLevel = this.fadeGoal;
	 this.fading = nil;
      end
   else 
      if (newLevel > this.fadeGoal) then
	 newLevel = this.fadeGoal;
	 this.fading = nil;
      end
   end
   this:SetAlpha(newLevel);
   this.curAlpha = newLevel;
end

local function ActivateFade(frame)
   if (not frame:IsVisible()) then
      frame:SetAlpha(frame.fadeGoal);
      frame.curAlpha = frame.fadeGoal;
      frame.fadeGoal = nil;
      frame.fading = nil;
      return;
   end
   if (not frame.fadeStart) then
      frame.fadeStart = GetTime();
   end
   if (not frame.curAlpha) then
      frame.curAlpha = frame:GetAlpha();
   end
   -- DEBUG("Starting fade " .. frame.fadeStart .. " " .. frame.fading .. " " .. frame.fadeGoal);
   frame:SetScript("OnUpdate", DoFade);
end


function METHODS:FadeSet(level) 
   self:SetAlpha(level);
   self.curAlpha = level;
   self.fading = nil;
end

function METHODS:FadeTo(level, time) 
   if (level > 1) then
      level = 1;
   elseif (level < 0) then
      level = 0;
   end
   if (time <= 0) then
      self:FadeSet(level);
      return;
   end

   local rate = 1.0 / time;
   local oldFading = self.fading;

   if (level == self.curAlpha) then
      self.fading = nil;
      return;
   end

   if (oldFading) then
      if (level < self.curAlpha) then
	 if (rate > -self.fading) then
	    self.fading = -rate;
	 end
      else
	 if (rate > self.fading) then
	    self.fading = rate;
	 end
      end
   else
      if (level < self.curAlpha) then
	 self.fading = -rate;
      else
	 self.fading = rate;
      end
   end

   self.fadeGoal = level;
   if (not oldFading) then
      ActivateFade(self);
   end
end

function METHODS:ApplyConfig(configSet, nonSticky, forceChange)
   local newVal,curVal;
   local anyChange = forceChange;
   for k,v in pairs(RINGCONFIG_STANDARD) do
      newVal = configSet[k] or v;
      curVal = self[k];
      if (curVal ~= newVal) then
	 self[k] = newVal;
	 anyChange = true;
      end
   end
   if (nonSticky) then
      self.currentConfig = nil;
   else
      self.currentConfig = configSet;
   end
   if (anyChange) then
      -- Check visibility, if it didn't change then update health and
      -- mana.
      self:FadeModelUpdate(true);
      self:NotifyInfoUpdate();
   end
end

function METHODS:ApplyScaffoldAnchor(scaffold, anchor, x, y, scale, spin)
   local parName = self:GetParent():GetName();

   self.scaffold = scaffold;

   if (x) then
      scale = scale / 100;
      self.noAnchor = false;
      self:SetScale(scale);
      self:ClearAllPoints();
      self:SetPoint("CENTER", parName, "CENTER", x/scale, y/scale);

      local refresh = nil;

      if (self.rotation ~= spin) then
	 self.rotation = spin;
	 self.requiresRefresh = true; -- Force a refresh regardless
	 refresh = true;
      end

      if (not self.visible) then
	 -- We shouldn't BE visible!
	 return;
      end
   
      if (not self:IsShown()) then
	 self:Show();
      elseif (refresh) then
	 self:NotifyInfoUpdate();
      end
   else
      self.noAnchor = true;
      self:Hide();
   end
end

function METHODS:UpdateVisibility(needVisible)
   local curVis = self.visible;
   if ((curVis and needVisible) or not (curVis or needVisible)) then
      return;
   end

   if (needVisible) then
      self.visible = true;
      if (self.noAnchor) then
	 local scaffold = self.scaffold;
	 --DEBUG("Should ask scaffold for an anchor now....");
	 if (scaffold and scaffold:CanSetActive(self.unit)) then
	    scaffold:SetActive(self.unit, true);
	    scaffold:Update();
	 end
	 return;
      end
      --DEBUG("Making <" .. self.unit .. "> ring visible");
      self:Show();
      return;
   end

   -- Otherwise, shouldn't be visible
   self.visible = false;
   --DEBUG("Making <" .. self.unit .. "> ring invisible");
   self:Hide();
   -- DEBUG("Maybe ask scaffold to hide anchor...");
   local scaffold = self.scaffold;
   if (scaffold and scaffold:CanSetActive(self.unit)) then
      scaffold:SetActive(self.unit, false);
      scaffold:Update();
   end
end

function METHODS:SetInfoModel(model, isDefault, forceNotify)
   if (isDefault) then
      self.defaultModel = model;
   end
   if (not model) then
      model = self.defaultModel;
   end
   if (self.infoModel == model) then
      if (forceNotify) then
	 self:NotifyInfoUpdate(model);
      end
      return;
   end

   -- Model change!
   self.infoModel = model;
   self.instantFade = true; -- Reset the fade model
   self.requiresRefresh = true; -- Force refresh of the model
   self:NotifyInfoUpdate(model);
end

function METHODS:NotifyInfoUpdate(model)
   if (not model) then
      model = self.infoModel;
   elseif (model ~= self.infoModel) then
      -- DEBUG("NotifyModelUpdate<" .. self.unit .. "> - IGNORED");
      return;
   end
   if (not model) then
      if (self.visible) then
	 self:UpdateVisibility(false);
      end
      return;
   end

   --DEBUG("NotifyModelUpdate<" .. self.unit .. ">");
   local force = model.force;
   if (force) then
      -- Less frequent, defer to another place
      self:DoForceInfoUpdate(model);
      return;
   end

   if (not model.shown) then
      -- If the ring isn't sure if it's hidden, then hide it for sure!
      if (self.visible) then
	 self:UpdateVisibility(false);
      end
      return;
   end
   
   -- At this point the ring SHOULD be visible, if it's not then
   -- stop here, but do something about it
   if (not (self.visible and self:IsVisible())) then
      -- Note that we need to refresh everything when the frame
      -- shows up.
      self.requiresRefresh = true;
      self:UpdateVisibility(true);
      return;
   end
   
   local needRefresh = self.requiresRefresh;
   if (needRefresh) then self.requiresRefresh = nil; end

   local health = model.healthPct or 0;
   local healthChange = needRefresh;
   if (self.visHealth ~= health) then
      self.visHealth = health;
      healthChange = true;
   end

   local mana = model.manaPct or 0;
   local manaMax = model.manaMax or 0;
   local manaChange = needRefresh;
   if ((self.visMana ~= mana) or (self.visManaMax ~= manaMax)) then
      self.visMana = mana;
      self.visManaMax = manaMax;
      manaChange = true;
   end
   
   local powerType = model.powerType or 0;
   local powerTypeChange = needRefresh;
   if (self.visPowerType ~= powerType) then
      self.visPowerType = powerType;
      powerTypeChange = true;
   end

   local name = model.unitName or "";
   local nameChange = needRefresh;
   if (self.visName ~= name) then
      nameChange = true;
      self.visName = name;
   end

   local alpha = model.alpha;
   local alphaChange = needRefresh;
   if (self.visAlpha ~= alpha) then
      alphaChange = true;
      self.visAlpha = alpha;
   end

   if (healthChange) then
      local r,g,b = 0,1,0;
      if (self.isTarget) then
	 r,g,b = 0.0,0.25,0.0;
      end
      if (self.smoothColor) then
	 if (value < 0.5) then
	    r = (1.0 - value) * 2;
	    g = 1.0;
	 else
	    r = 1.0;
	    g = value * 2;
	 end
      end
      --if (isGrey) then
      --local y = r*0.5+g*0.3+b*0.2;
      --r=y; g=y; b=y;
   --end
      self.healthFrame:CallTextureMethod("SetVertexColor", r, g, b, 1);
      self.healthFrame:SetAngle(self.ringLayout:GetHealthAngle(health, 
							       self.rotation));
   end
   if (powerTypeChange) then
      local info = ManaBarColor[powerType] or ManaBarColor[0];
      if (info) then
	 local r,g,b;
	 if (self.isTarget) then
	    r = info.r/4; g=info.g/4; b=info.b/4;
	 else
	    r = info.r;   g=info.g;   b=info.b;
	 end
	 --if (isGrey) then
	 --local y = r*0.5+g*0.3+b*0.2;
	 --r=y; g=y; b=y;
	 --end
	 self.manaFrame:CallTextureMethod("SetVertexColor", r, g, b, 1);
      end
   end
   if (manaChange) then
      self.manaFrame:SetAngle(self.ringLayout:GetManaAngle(mana,
							   self.rotation));
   end
   if (nameChange) then
      self:UpdateName(name);
   end
   local instant = self.instantFade;
   if (manaChange or healthChange or alphaChange or instant) then
      self:FadeModelUpdate(instant);
      self.instantFade = nil;
   end
end

function METHODS:UpdateName(name)
   if ((not name) or (name == "")) then
      self.nameFrame:Hide();
   elseif (name == true) then
      self.nameFrame:SetText("<"..self.unit..">");
      self.nameFrame:Show();
   else
      self.nameFrame:SetText(name);
      self.nameFrame:Show();
   end
end

function METHODS:DoForceInfoUpdate(model)
   local force = model.force;
   if (force == "grey") then
      if (not self.visible) then
	 self.requiresRefresh = true;
	 self:UpdateVisibility(true);
	 return;
      end

      self.healthFrame:CallTextureMethod("SetVertexColor", 0.4, 0.4, 0.4, 1);
      self.manaFrame:CallTextureMethod("SetVertexColor", 0.4, 0.4, 0.4, 1);

      self.healthFrame:SetAngle(self.ringLayout:GetHealthAngle(1, 
							       self.rotation));
      self.manaFrame:SetAngle(self.ringLayout:GetManaAngle(1,
							   self.rotation));
      self:FadeSet(model.alpha or 1);
      self.visName = model.unitName;
      self:UpdateName(model.unitName);
      return;
   end

   DEBUG("DoForceInfoUpdate<" .. self.unit .. "> " .. model.force);
end

function StatRingsUnitFrame_OnLoad(this, unit, hideName)
   local name = this:GetName();
   this.unit = unit;
   this.healthFrame = getglobal(name .. "Health");
   this.manaFrame = getglobal(name .. "Mana");
   this.nameFrame = getglobal(name .. "Name");

   this.hideName = hideName;
   this.rotation = 0;

   this.isTarget = (unit == "target");

   local newMeta = {};
   local oldMeta = getmetatable(this);
   for k,v in pairs(oldMeta) do newMeta[k] = v; end
   function newMeta.__index(table, index) 
      return oldMeta.__index(table,index) or METHODS[index];
   end
   setmetatable(this, newMeta);
   
   this.manaFrame:SetPoint("CENTER", name, "CENTER", 0, 0);
   this.healthFrame:SetPoint("CENTER", name, "CENTER", 0, 0);
   
   this.FadeModelUpdate = StatRingsFadeModel_Update;

   this.unitModel = StatRingsUnitModel_New(this);
   this.unitModel:Initialize(this, unit);
   this.unitModel:CreateOnEvent();

   this.ringLayout = RINGLAYOUT_SPLIT;
   this.ringLayout:SetupRings(this);

   -- Apply default configuration but dont link to it
   this:ApplyConfig(RINGCONFIG_STANDARD, true, true);
   this:SetInfoModel(this.unitModel, true);
end

function StatRingsUnitFrame_ComboOnEvent(this, event)
   this:SetAngle(130 * GetComboPoints() / MAX_COMBO_POINTS);
end

